# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Causal Manager class."""
import numpy as np
import pandas as pd

from econml.solutions.causal_analysis import CausalAnalysis

from responsibleai._config.base_config import BaseConfig
from responsibleai._internal.constants import ManagerNames
from responsibleai._managers.base_manager import BaseManager
from responsibleai.exceptions import (
    UserConfigValidationException, DuplicateManagerConfigException)
from responsibleai.modelanalysis.constants import ModelTask
from responsibleai._interfaces import CausalData


class CausalConstants:
    # Model types
    AUTOML = 'automl'
    LINEAR = 'linear'


class CausalConfig(BaseConfig):
    def __init__(
        self,
        treatment_features=None,
        nuisance_model=None,
        heterogeneity_model=None,
        alpha=0.05,
        max_cat_expansion=5,
        treatment_cost=0,
        min_tree_leaf_samples=2,
        max_tree_depth=3,
    ):
        super().__init__()
        self.treatment_features = treatment_features
        self.nuisance_model = nuisance_model
        self.heterogeneity_model = heterogeneity_model
        self.alpha = alpha
        self.max_cat_expansion = max_cat_expansion
        self.treatment_cost = treatment_cost
        self.min_tree_leaf_samples = min_tree_leaf_samples
        self.max_tree_depth = max_tree_depth

        # Outputs
        self.causal_analysis = None
        self.global_effects = None
        self.local_effects = None
        self.policies = None

    def __eq__(self, other):
        return all([
            np.array_equal(self.treatment_features,
                           other.treatment_features),
            self.nuisance_model == other.nuisance_model,
            self.heterogeneity_model == other.heterogeneity_model,
            self.alpha == other.alpha,
            self.max_cat_expansion == other.max_cat_expansion,
            self.treatment_cost == other.treatment_cost,
            self.min_tree_leaf_samples == other.min_tree_leaf_samples,
            self.max_tree_depth == other.max_tree_depth,
        ])

    def __repr__(self):
        return ("CausalConfig("
                f"treatment_features={self.treatment_features}, "
                f"nuisance_model={self.nuisance_model}, "
                f"heterogeneity_model={self.heterogeneity_model}, "
                f"alpha={self.alpha}, "
                f"max_cat_expansion={self.max_cat_expansion}, "
                f"treatment_cost={self.treatment_cost}, "
                f"min_tree_leaf_samples={self.min_tree_leaf_samples}, "
                f"max_tree_depth={self.max_tree_depth})")

    def to_result(self):
        return {
            'causal_analysis': self.causal_analysis,
            'global_effects': self.global_effects,
            'local_effects': self.local_effects,
            'policies': self.policies,
        }


class CausalManager(BaseManager):
    def __init__(self, train, test, target_column, task_type,
                 categorical_features):
        """Construct a CausalManager for generating causal analyses
           from a dataset.

        :param train: Dataset on which to compute global causal effects
                     (#samples x #features).
        :type train: pandas.DataFrame
        :param test: Dataset on which to compute local causal effects
                     (#samples x #features).
        :type test: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param task_type: Task type is either 'classification/regression'
        :type task_type: str
        :param categorical_features: All categorical feature names.
        :type categorical_features: list
        """
        self._train = train
        self._test = test
        self._target_column = target_column
        self._task_type = task_type
        self._categorical_features = categorical_features
        self._causal_config_list = []

    def add(
        self,
        treatment_features=None,
        nuisance_model=CausalConstants.AUTOML,
        heterogeneity_model=None,
        alpha=0.05,
        max_cat_expansion=5,
        treatment_cost=0,
        min_tree_leaf_samples=2,
        max_tree_depth=3,
    ):
        """Add a causal configuration to be computed later.
        :param treatment_features: All treatment feature names.
        :type treatment_features: list
        :param nuisance_model: Model type to use for nuisance estimation.
        :type nuisance_model: str
        :param heterogeneity_model: Model type to use for
                                    treatment effect heterogeneity.
        :type heterogeneity_model: str
        :param alpha: Confidence level of confidence intervals.
        :type alpha: float
        :param max_cat_expansion: Maximum expansion for categorical features.
        :type max_cat_expansion: int
        :param treatment_cost: Cost to treat one individual or
                               per-individual costs as an array.
        :type treatment_cost: float or array
        :param min_tree_leaf_samples: Minimum number of samples per leaf
                                      in policy tree.
        :type min_tree_leaf_samples: int
        :param max_tree_depth: Maximum depth of policy tree.
        :type max_tree_depth: int
        """
        causal_config = CausalConfig(
            treatment_features=treatment_features,
            nuisance_model=nuisance_model,
            heterogeneity_model=heterogeneity_model,
            alpha=alpha,
            max_cat_expansion=max_cat_expansion,
            treatment_cost=treatment_cost,
            min_tree_leaf_samples=min_tree_leaf_samples,
            max_tree_depth=max_tree_depth)

        if causal_config.is_duplicate(self._causal_config_list):
            raise DuplicateManagerConfigException(
                "Duplicate causal configuration detected.")

        self._causal_config_list.append(causal_config)

    def compute(self):
        """Computes the causal effects by running the causal configuration."""
        for config in self._causal_config_list:
            if config.is_computed:
                continue

            config.is_computed = True
            if config.nuisance_model not in [CausalConstants.AUTOML,
                                             CausalConstants.LINEAR]:
                message = (f"nuisance_model should be one of "
                           f"['{CausalConstants.AUTOML}', "
                           f"'{CausalConstants.LINEAR}'], "
                           f"got {config.nuisance_model}")
                raise UserConfigValidationException(message)

            is_classification = self._task_type == ModelTask.CLASSIFICATION
            X = pd.concat([self._train, self._test], ignore_index=True)\
                .drop([self._target_column], axis=1)
            y = pd.concat([self._train, self._test], ignore_index=True)[
                self._target_column].values.ravel()

            analysis = CausalAnalysis(
                X.columns.values.tolist(),
                self._categorical_features,
                heterogeneity_inds=config.treatment_features,
                classification=is_classification,
                nuisance_models=config.nuisance_model,
                upper_bound_on_cat_expansion=config.max_cat_expansion,
                n_jobs=-1)
            analysis.fit(X, y)

            config.causal_analysis = analysis

            config.global_effects = analysis.global_causal_effect(
                alpha=0.05)
            X_test = self._test.drop([self._target_column], axis=1)
            config.local_effects = analysis.local_causal_effect(
                X_test)

            config.policies = []
            if config.treatment_features is not None:
                for treatment_feature in config.treatment_features:
                    local_policies = analysis._individualized_policy_dict(
                        X_test, treatment_feature,
                        treatment_costs=config.treatment_cost,
                        alpha=config.alpha)

                    policy_tree, recommended_gains, treatment_gains = \
                        analysis._policy_tree_output(
                            X_test, treatment_feature,
                            treatment_costs=config.treatment_cost,
                            max_depth=config.max_tree_depth,
                            min_samples_leaf=config.min_tree_leaf_samples,
                            alpha=config.alpha)

                    policy = {
                        'local_policies': local_policies,
                        'policy_tree': policy_tree,
                        'policy_gains': {
                            'recommended_policy_gains': recommended_gains,
                            'treatment_gains': treatment_gains,
                        }
                    }
                    config.policies.append(policy)

    def get(self):
        """Get the computed causal effects."""
        results = []
        for config in self._causal_config_list:
            if config.is_computed:
                results.append(config.to_result())
        return results

    def list(self):
        pass

    def get_data(self):
        """Get causal data

        :return: A array of CausalData.
        :rtype: List[CausalData]
        """
        return [self._get_causal(i)
                for i in self.get()]

    def _get_causal(self, causal):
        causal_data = CausalData()
        causal_data.globalCausalEffects = causal["global_effects"]\
            .reset_index().to_dict(orient="records")
        causal_data.localCausalEffects = causal["local_effects"]\
            .groupby("sample").apply(
                lambda x: x.reset_index().to_dict(
                    orient='records')).values
        return causal_data

    @property
    def name(self):
        """Get the name of the causal manager.
        :return: The name of the causal manager.
        :rtype: str
        """
        return ManagerNames.CAUSAL

    def _save(self, path):
        pass

    @staticmethod
    def _load(path, model_analysis):
        pass



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from enum import Enum
from typing import List


class Dataset:
    predictedY: List
    features: List[List[int]]
    featureNames: List[str]
    probabilityY: List
    trueY: List
    classNames: List[str]


class BoundedCoordinates:
    type: str
    names: List
    scores: List
    scores_range: List
    upper_bounds: List
    lower_bounds: List


class EBMGlobalExplanation:
    feature_list: List[BoundedCoordinates]


class FeatureImportance:
    scores: List
    intercept: float
    featureNames: List[str]


class PrecomputedExplanations:
    localFeatureImportance: FeatureImportance
    globalFeatureImportance: FeatureImportance
    ebmGlobalExplanation: EBMGlobalExplanation
    customVis: str


class ModelClass(str, Enum):
    TREE = 'Tree'
    EBM = 'EBM'
    BLACKBOX = 'blackbox'


class ModelMethod(str, Enum):
    CLASSIFIER = 'classifier'
    REGRESSOR = 'regressor'


class ModelExplanationData:
    modelClass: ModelClass
    method: ModelMethod
    predictedY: List
    probabilityY: List[List[float]]
    explanationMethod: str
    precomputedExplanations: PrecomputedExplanations


class ErrorAnalysisData:
    maxDepth: int
    numLeaves: int


class CausalMetric:
    ci_lower: float
    ci_upper: float
    feature: str
    p_value: float
    point: float
    stderr: float
    zstat: float


class CausalData:
    globalCausalEffects: List[CausalMetric]
    localCausalEffects: List[List[CausalMetric]]


class CounterfactualData:
    cfsList: List[List[List[float]]]
    featureNames: List[float]


class ModelAnalysisData:
    dataset: Dataset
    modelExplanationData: List[ModelExplanationData]
    causalAnalysisData: List[CausalData]
    counterfactualData: List[CounterfactualData]
    errorAnalysisConfig: List[ErrorAnalysisData]



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pytest

import pandas as pd

from responsibleai.exceptions import (
    DuplicateManagerConfigException,
    UserConfigValidationException)

from econml.solutions.causal_analysis._causal_analysis import CausalAnalysis


EFFECTS_ATTRIBUTES = [
    # 'raw_name',
    # 'name',
    # 'cat',
    # 'type',
    # 'version',
    # 'causal_computation_type',
    # 'confounding_interval',
    # 'init_args',
    'point',
    'stderr',
    'zstat',
    'ci_lower',
    'ci_upper',
    'p_value',
    # 'view',
]

LOCAL_POLICY_ATTRIBUTES = [
    'Treatment',
    'Effect of treatment',
    'Effect of treatment lower bound',
    'Effect of treatment upper bound',
    'Current treatment'
]


def validate_causal(model_analysis, data, target_column,
                    treatment_features, max_cat_expansion):
    # Add the first configuration
    model_analysis.causal.add(nuisance_model='automl',
                              treatment_features=treatment_features,
                              max_cat_expansion=max_cat_expansion)
    model_analysis.causal.compute()
    results = model_analysis.causal.get()
    assert results is not None
    assert isinstance(results, list)
    assert len(results) == 1
    causal_results = results[0]
    _check_causal_results(causal_results)

    # Add a duplicate configuration
    message = "Duplicate causal configuration detected."
    with pytest.raises(DuplicateManagerConfigException, match=message):
        model_analysis.causal.add(nuisance_model='automl',
                                  treatment_features=treatment_features,
                                  max_cat_expansion=max_cat_expansion)

    # Add the second configuration
    model_analysis.causal.add(nuisance_model='linear')
    model_analysis.causal.compute()
    results = model_analysis.causal.get()
    assert results is not None
    assert isinstance(results, list)
    assert len(results) == 2

    # Add a bad configuration
    model_analysis.causal.add(nuisance_model='fake_model')
    with pytest.raises(UserConfigValidationException):
        model_analysis.causal.compute()


def _check_causal_results(causal_results):
    assert isinstance(causal_results, dict)
    assert len(causal_results) == 4

    assert 'causal_analysis' in causal_results
    _check_causal_analysis(causal_results['causal_analysis'])
    assert 'global_effects' in causal_results
    _check_global_effects(causal_results['global_effects'])
    assert 'local_effects' in causal_results
    _check_local_effects(causal_results['local_effects'])
    assert 'policies' in causal_results
    _check_policies(causal_results['policies'])


def _check_causal_analysis(causal_analysis):
    assert isinstance(causal_analysis, CausalAnalysis)


def _check_global_effects(global_effects):
    assert isinstance(global_effects, pd.DataFrame)
    for attribute in EFFECTS_ATTRIBUTES:
        assert attribute in global_effects.columns.to_list()


def _check_local_effects(local_effects):
    assert isinstance(local_effects, pd.DataFrame)
    for attribute in EFFECTS_ATTRIBUTES:
        assert attribute in local_effects.columns.to_list()


def _check_policies(policies):
    assert isinstance(policies, list)
    for policy in policies:
        _check_policy(policy)


def _check_policy(policy):
    assert isinstance(policy, dict)

    assert 'local_policies' in policy
    local_policies = policy['local_policies']
    assert isinstance(local_policies, dict)
    for attribute in LOCAL_POLICY_ATTRIBUTES:
        assert attribute in local_policies

    assert 'policy_tree' in policy
    policy_tree = policy['policy_tree']
    _check_policy_tree(policy_tree)

    assert 'policy_gains' in policy
    policy_gains = policy['policy_gains']
    assert isinstance(policy_gains, dict)

    assert 'recommended_policy_gains' in policy_gains
    recommended_policy_gains = policy_gains['recommended_policy_gains']
    assert isinstance(recommended_policy_gains, float)

    assert 'treatment_gains' in policy_gains
    treatment_gains = policy_gains['treatment_gains']
    assert isinstance(treatment_gains, list)
    assert all(isinstance(v, float) for v in treatment_gains)


def _check_policy_tree(policy_tree):
    assert 'leaf' in policy_tree
    leaf = policy_tree['leaf']

    if leaf:
        assert 'n_samples' in policy_tree
        assert 'treatment' in policy_tree
    else:
        assert 'feature' in policy_tree
        assert isinstance(policy_tree['feature'], str)

        assert 'threshold' in policy_tree
        assert isinstance(policy_tree['threshold'], float)

        assert 'left' in policy_tree
        assert 'right' in policy_tree
        _check_policy_tree(policy_tree['left'])
        _check_policy_tree(policy_tree['right'])



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pytest
from responsibleai import ModelTask
from responsibleai.exceptions import (
    DuplicateManagerConfigException, UserConfigValidationException
)


def validate_counterfactual(cf_analyzer, X_train, target_column,
                            desired_class=None, desired_range=None):
    continuous_features = list(set(X_train.columns) - set([target_column]))

    # Add the first configuration
    cf_analyzer.counterfactual.add(continuous_features=continuous_features,
                                   total_CFs=10,
                                   method='random',
                                   desired_class=desired_class,
                                   desired_range=desired_range)
    cf_analyzer.counterfactual.compute()
    assert cf_analyzer.counterfactual.get() is not None
    assert isinstance(cf_analyzer.counterfactual.get(), list)
    assert len(cf_analyzer.counterfactual.get()) == 1

    # Add a duplicate configuration
    with pytest.raises(DuplicateManagerConfigException):
        cf_analyzer.counterfactual.add(continuous_features=continuous_features,
                                       total_CFs=10,
                                       method='random',
                                       desired_class=desired_class,
                                       desired_range=desired_range)

    # Add the second configuration
    cf_analyzer.counterfactual.add(continuous_features=continuous_features,
                                   total_CFs=1,
                                   method='random',
                                   desired_class=desired_class,
                                   desired_range=desired_range)
    cf_analyzer.counterfactual.compute()
    assert cf_analyzer.counterfactual.get() is not None
    assert isinstance(cf_analyzer.counterfactual.get(), list)
    assert len(cf_analyzer.counterfactual.get()) == 2

    # Add a bad configuration
    cf_analyzer.counterfactual.add(continuous_features=continuous_features,
                                   total_CFs=-1,
                                   method='random',
                                   desired_class=desired_class,
                                   desired_range=desired_range)
    cf_analyzer.counterfactual.compute()
    assert cf_analyzer.counterfactual.get() is not None
    assert isinstance(cf_analyzer.counterfactual.get(), list)
    assert len(cf_analyzer.counterfactual.get()) == 2
    assert len(cf_analyzer.counterfactual.get(failed_to_compute=True)) == 1

    task_type = cf_analyzer.task_type
    if task_type == ModelTask.REGRESSION:
        with pytest.raises(UserConfigValidationException):
            cf_analyzer.counterfactual.add(
                continuous_features=continuous_features,
                total_CFs=10,
                method='random',
                desired_range=None)
    else:
        with pytest.raises(UserConfigValidationException):
            cf_analyzer.counterfactual.add(
                continuous_features=continuous_features,
                total_CFs=10,
                method='random',
                desired_class=None)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
import pytest

from erroranalysis._internal.matrix_filter import (
    CATEGORY1, CATEGORY2, COUNT, FALSE_COUNT, MATRIX, VALUES)
from responsibleai.exceptions import DuplicateManagerConfigException

SIZE = 'size'
PARENTID = 'parentId'
ERROR = 'error'
ID = 'id'


def setup_error_analysis(model_analysis, add_ea=True):
    if add_ea:
        model_analysis.error_analysis.add()
        with pytest.raises(DuplicateManagerConfigException):
            model_analysis.error_analysis.add()
    model_analysis.error_analysis.compute()


def validate_error_analysis(model_analysis):
    reports = model_analysis.error_analysis.get()
    assert isinstance(reports, list)
    assert len(reports) == 1
    report = reports[0]
    json_matrix = report.json_matrix

    ea_x_train = model_analysis.error_analysis._train
    ea_y_train = model_analysis.error_analysis._y_train

    expected_count = len(ea_x_train)
    if json_matrix is not None:
        predictions = model_analysis.model.predict(ea_x_train)
        expected_false_count = sum(predictions != ea_y_train)
        validate_matrix(json_matrix, expected_count, expected_false_count)

    json_tree = report.json_tree
    validate_tree(json_tree, expected_count)


def validate_matrix(json_matrix, exp_total_count, exp_total_false_count):
    assert MATRIX in json_matrix
    assert CATEGORY1 in json_matrix
    assert CATEGORY2 in json_matrix
    num_cat1 = len(json_matrix[CATEGORY1][VALUES])
    num_cat2 = len(json_matrix[CATEGORY2][VALUES])
    assert len(json_matrix[MATRIX]) == num_cat1
    assert len(json_matrix[MATRIX][0]) == num_cat2
    # take sum of count, false count
    total_count = 0
    total_false_count = 0
    for i in range(num_cat1):
        for j in range(num_cat2):
            total_count += json_matrix[MATRIX][i][j][COUNT]
            total_false_count += json_matrix[MATRIX][i][j][FALSE_COUNT]
    assert exp_total_count == total_count
    assert exp_total_false_count == total_false_count


def validate_tree(json_tree, expected_count):
    assert json_tree is not None
    assert len(json_tree) > 0
    assert ERROR in json_tree[0]
    assert ID in json_tree[0]
    assert PARENTID in json_tree[0]
    assert json_tree[0][PARENTID] is None
    assert SIZE in json_tree[0]
    assert json_tree[0][SIZE] == expected_count
